/**********************************************************************
    fepdos.c : Fep Bridge 1.33 (1996/11/13)

    Copyright (C) Yutaka Tamiya (tamy@flab.fujitsu.co.jp)

    I stole some codes from:
        "fepctrl.c 1.2" 1990/11/25
        General purpose Japanese FEP control routines for MS-DOS.
                Written by Junn Ohta <ohta@src.ricoh.co.jp>
**********************************************************************/
#include <dos.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <memory.h>
#include <fcntl.h>
#include <signal.h>

const char * const Version =	"fepdos.exe ver. 3.00 (Fep Bridge ver. 1.33)\n"
				"\tby Yutaka Tamiya (tamy@flab.fujitsu.co.jp)";

/*
    FIFO's for comunication between Dos and Unix

    FIFO_DEFAULT_DIR     = Default location of FIFO's
    FIFO_FEP_IN  = Dos's input FIFO:  Unix  -> Dos
    FIFO_FEP_OUT = Dos's output FIFO: Dos -> Unix
    FIFO_FEP_MSG = Dos's message FIFO:Dos -> Unix
*/

#define FIFO_DEFAULT_DIR        "e:\\"
#define FIFO_FEP_IN     "fepin"
#define FIFO_FEP_OUT    "fepout"
#define FIFO_FEP_MSG    "fepmsg"

/*
    Packet ID
*/
typedef enum {
    FO_Info	= 1,
    FO_Ack	= 2,	/* This program never send this packet */
    FO_Result	= 5,
    FO_Alive	= 3,
    FO_Dead	= 4,	/* This program never send this packet */
    FO_UserList	= 6,	/* This program never send this packet */
} FepOutPacketID;

typedef enum {
    FI_Info	= 1,
    FI_Auth	= 3,	/* This program never get this packet */
    FI_Reset	= 2,
    FI_Conv1	= 4,
    FI_ConvN	= 6,
    FI_Conv	= FI_Conv1,
    FI_Stat	= 5,
    FI_Kill	= 7,	/* ^G */
    FI_UserList	= 8,	/* This program never get this packet */
    FI_Nop	= 127,
} FepInPacketID;

typedef struct {
    char	scan;
    char	ascii;
    char	shift;
} KeyCode;

/*
    Global Variables
*/
static int	FepIn = -1;	/* File handle of FIFO_FEP_IN   */
static FILE	*FepOut = 0;	/* File handle of FIFO_FEP_OUT  */
static FILE	*FepMsg;	/* File handle of FIFO_FEP_MSG  */
static int	DebugFlag;	/* Debug messages to FIFO_FEP_DEBUG */
static int	HaltFlag = 0;	/* Halt the system on exit	*/

#define GET_IOC_DATA    0x4400  /* Get IOCTL Data               */
#define RCV_IOC_CHAR    0x4402  /* Receive IOCTL Character      */
#define GET_CBRK        0x3300  /* Get CBREAK (CTRL-C Check)    */
#define SET_CBRK        0x3301  /* Set CBREAK (CTRL-C Check)    */
#define GET_VECTOR      0x35    /* Get Interrupt Vector         */
#define OPEN_HANDLE     0x3d    /* Open Handle                  */
#define CLOSE_HANDLE    0x3e    /* Close Handle                 */
#define DEVICE          0x80    /* handle is not a file         */

/* bit mask of shift status */
#define	VK_SHIFT_MASK	0x01
#define	VK_CONTROL_MASK	0x02
#define	VK_ALT_MASK	0x04

/*
 * MS-KANJI Application Interface access routines
 */

typedef struct {
    short       wLevel;                 /* API support level */
    char        rgchName[8];            /* name of system */
    char        rgchVersion[4];         /* version # of system */
    char        rgchRevision[4];        /* revision # of system */
    char        Reserved[14];           /* reserved */
} Kkname;

typedef struct {
    short       wType;                  /* display switch */
    short       wScan;                  /* key scan code */
    short       wAscii;                 /* ascii code */
    short       wShift;                 /* shift key status */
    short       wExShift;               /* extended shift key status */
    short       cchResult;              /* size of confirmed string */
    char far    *lpchResult;            /* ptr to confirmed string */
    short       cchMode;                /* size of mode display str */
    char far    *lpchMode;              /* ptr to mode display str */
    char far    *lpattrMode;            /* ptr to attribute of it */
    short       cchSystem;              /* size of system string */
    char far    *lpchSystem;            /* ptr to system string */
    char far    *lpattrSystem;          /* ptr to attribute of it */
    short       cchBuf;                 /* size of unconfirmed str */
    char far    *lpchBuf;               /* ptr to unconfirmed str */
    char far    *lpattrBuf;             /* ptr to attribute of it */
    short       cchBufCursor;           /* cursor position in it */
    char        Reserved[34];           /* reserved */
} Dataparm;

typedef struct {
    short       wFunc;                  /* function # */
    short       wMode;                  /* reference or set */
    Kkname far *lpKkname;               /* translation system */
    Dataparm far *lpDataparm;           /* data parameter */
    char        Reserved[4];            /* reserved */
} Funcparm;

/*
 * function # for wFunc
 */
#define KKAsk   1
#define KKOpen  2
#define KKClose 3
#define KKInOut 4
#define KKMode  5

/*
 * mode bit for wMode
 */
#define KM_GET  0x0000
#define KM_SET  0x8000
#define KM_OFF  0x0001
#define KM_ON   0x0002
#define KM_SYS  0x0004
#define KM_ECHO 0x0008


/*
   KKfunc(), the MS-KANJI API I/F function, is a "Pascal" type function, i.e.:
    - All arguments of KKfunc() are stored on the stack, not in the registers.
    - KKfunc() retreives the stack pointer before returning to the caller.
   In the case of MS-C, the function declaration must be:

	static int (far pascal *KKfunc)();

   LSI C-86 does not support "Pascal" type functions.
   So, I did some tricks:
    - The arg declaration of "..." makes arguments stored on the stack.
    - After calling KKfunc(), adjust the stack pointer with inline asm code.
      (See callKKFunc() )
*/
static int (far * KKfunc)(Funcparm far *lpFuncparm, ...);

/*
    htons() : Transform i86 short int to short int of the network byte order
    The following inline asm code (for LSI C-86) means:
	mov ax, c	; arg "c" is stored to AX.
	xchg ah, al	; swap high and low byte
			; the return value is in AX.
*/
static unsigned short _asm_htons(char *asm_statement, unsigned short hostShort);
#define htons(c)       _asm_htons("\n\tXCHG\tAH, AL\n", c)
#define ntohs(c)       htons(c)

static Funcparm Funcbuf;
static Kkname Kkbuf;
static Dataparm Databuf;

/* near pointers of strings/attributes */
static char * LpchResult;
static char * LpchMode;
static char * LpattrMode;
static char * LpchSystem;
static char * LpattrSystem;
static char * LpchBuf;
static char * LpattrBuf;

/* Maximum length of Result, Mode, System, and Buf */
static short  CchMaxResult;
static short  CchMaxMode;
static short  CchMaxSystem;
static short  CchMaxBuf;

/* Store initial strings of Mode and System */
static short  CchInitMode;
static short  CchInitSystem;
static char * LpchInitMode;
static char * LpattrInitMode;
static char * LpchInitSystem;
static char * LpattrInitSystem;

/*
    sub functions
*/
static int      dopen(const char *dname);
static void     dclose(int fd);
static int      callKKFunc(Funcparm far * bufp);
static void	sigHandlerFepBrdg(int sig);
static int      initFepBrdg(int ac, const char *const av[]);
static void     terminateFepBrdg(void);
static void	haltTheSystem(void);
static void     usage(void);

static int      msknjFind(void);
static int      msknjOpen(void);
static int      msknjClose(void);
static int      msknjInOut(const KeyCode *keyCode);
static int      msknjGetMode(void);
static int      msknjSetMode(int on);

static FepInPacketID    recvPacketID(void);
static void     recvConvKey(KeyCode *keyCode);
static void     sendInfo(void);
static void     sendResult(int rturnValue);
static void     sendStat(void);

int
main(int ac, char *av[])
{
    int status = 0;
    int rt = 0;
    short n;
    int i;
    FepInPacketID       packetID;
    KeyCode     keyCode;

    if (initFepBrdg(ac, av) == 0) {
        status = 1;
	goto Exit;
    }

    /* Do sigHandlerFepBrdg() before eixt */
    signal(SIGABRT, sigHandlerFepBrdg);
    signal(SIGFPE,  sigHandlerFepBrdg);
    signal(SIGILL,  sigHandlerFepBrdg);
    signal(SIGINT,  sigHandlerFepBrdg);
    signal(SIGSEGV, sigHandlerFepBrdg);
    signal(SIGTERM, sigHandlerFepBrdg);

    /* send FO_Alive to "fepbrdg" that I'm ready. (from ver. 3.00) */
    sendStat();

    /* this is a Finite State Machine with 2 states */
    while (1) {
        packetID = recvPacketID();
        switch(packetID){
        case FI_Info:
            sendInfo();
            continue;
        case FI_Reset:
            msknjSetMode(0);
            msknjSetMode(1);
            sendResult(0);
            continue;
        case FI_Conv1:
            recvConvKey(&keyCode);
            rt = msknjInOut(&keyCode);
	    /* rt = 1, if the key event should be sent back to the client */
	    /*      0, FEP accepts the key event normally		  */
            sendResult(rt);
            continue;
        case FI_ConvN:
	    rt = 0;
	    read(FepIn, &n, 2);	/* Read # of keys to be read */
	    for (i = ntohs(n); i > 0; i--) {
		recvConvKey(&keyCode);
		rt |= msknjInOut(&keyCode);
		/* rt = 1, if the key event should be sent back to the client*/
		/*      0, FEP accepts the key event normally		  */
	    }
	    sendResult(rt);
            continue;
        case FI_Kill:
            goto Exit;
        case FI_Stat:
            sendStat();
            continue;
        }
    }

Exit:
    terminateFepBrdg();
    if (HaltFlag) haltTheSystem();

    return status;
}

/*
    Receive Packet from Unix to Dos
*/
static FepInPacketID
recvPacketID(void)
{
    char        c;

    /* wait until valid data comes into FepIn */
    while (read(FepIn, &c, 1) <= 0);
    if (DebugFlag) {
	fprintf(FepMsg, "Get Packet of %d\n", c);
	fflush(FepMsg);
    }
    return (FepInPacketID)c;
}

#include "keycode.c"

static void
recvConvKey(KeyCode *keyCode)
{
    unsigned char	buf[3];
    /* buf[0] = MS-Windows virtual key code, buf[1] = ascii, buf[2] = shift */

    read(FepIn, buf, 3);
    keyCode->ascii = buf[1];
    keyCode->shift = (buf[2] << 1);
    if (buf[0] == 0) {
	/* get scan code from ASCII code */
	keyCode->scan = AsciiToBiosKeyCode[keyCode->ascii];
    } else {
	if (buf[2] & VK_ALT_MASK) {
	    keyCode->scan = VirtualKeyCodeToBiosKeyCode[buf[0]].alt;
	} else if (buf[2] & VK_CONTROL_MASK) {
	    keyCode->scan = VirtualKeyCodeToBiosKeyCode[buf[0]].control;
	} else if (buf[2] & VK_SHIFT_MASK) {
	    keyCode->scan = VirtualKeyCodeToBiosKeyCode[buf[0]].shift;
	} else { /* no shift */
	    keyCode->scan = VirtualKeyCodeToBiosKeyCode[buf[0]].noShift;
	}
    }
    if (DebugFlag) {
        fprintf(FepMsg,"Get KeyCode %02x:%02x:%02x:%02x (%c) -->",
            buf[0], buf[1], buf[2], 0, buf[1]);
        fprintf(FepMsg," %d:%d\n", keyCode->scan, keyCode->ascii);
        fflush(FepMsg);
    }
}

/*
    Send Packet from Dos to Unix
*/

static void
sendInfo(void)
{
    unsigned short	nw;	/* short integer of the network byte order */
    const char	c = FO_Info;
    fwrite(&c, 1, 1, FepOut);
    /* send MS-KANJI API level */
    nw = htons(Kkbuf.wLevel);
    fwrite((char *)&nw, 2, 1, FepOut);
    /* send maximum string size of Result, Mode, System and Buf */
    nw = htons(CchMaxResult);
    fwrite((char *)&nw, 2, 1, FepOut);
    nw = htons(CchMaxMode);
    fwrite((char *)&nw, 2, 1, FepOut);
    nw = htons(CchMaxSystem);
    fwrite((char *)&nw, 2, 1, FepOut);
    nw = htons(CchMaxBuf);
    fwrite((char *)&nw, 2, 1, FepOut);
    nw = htons(CchInitMode);
    fwrite((char *)&nw, 2, 1, FepOut);
    nw = htons(CchInitSystem);
    fwrite((char *)&nw, 2, 1, FepOut);

    fwrite(LpchInitMode, 1, CchInitMode, FepOut);
    fwrite(LpattrInitMode, 1, CchInitMode, FepOut);
    fwrite(LpchInitSystem, 1, CchInitSystem, FepOut);
    fwrite(LpattrInitSystem, 1, CchInitSystem, FepOut);
    fflush(FepOut);

    if (DebugFlag) {
        fprintf(FepMsg,"sendInfo\n");
        fflush(FepMsg);
    }
}

static void
sendResult(int returnValue)
{
    char	c = FO_Result;
    unsigned short	nw;	/* short integer of the network byte order */

    fwrite(&c, 1, 1, FepOut);
    c = returnValue;
    /* return value of KKFunc() */
    fwrite(&c, 1, 1, FepOut);
    if (c == 0) {
	nw = htons(Databuf.cchResult);
	fwrite((char *)&nw, 2, 1, FepOut);
	nw = htons(Databuf.cchMode);
	fwrite((char *)&nw, 2, 1, FepOut);
	nw = htons(Databuf.cchSystem);
	fwrite((char *)&nw, 2, 1, FepOut);
	nw = htons(Databuf.cchBuf);
	fwrite((char *)&nw, 2, 1, FepOut);
	nw = htons(Databuf.cchBufCursor);
	fwrite((char *)&nw, 2, 1, FepOut);
	fwrite(LpchResult, 1, Databuf.cchResult, FepOut);
	fwrite(LpchMode, 1, Databuf.cchMode, FepOut);
	fwrite(LpattrMode, 1, Databuf.cchMode, FepOut);
	fwrite(LpchSystem, 1, Databuf.cchSystem, FepOut);
	fwrite(LpattrSystem, 1, Databuf.cchSystem, FepOut);
	fwrite(LpchBuf, 1, Databuf.cchBuf, FepOut);
	fwrite(LpattrBuf, 1, Databuf.cchBuf, FepOut);

	if (DebugFlag) {
	    fprintf(FepMsg, "Result: %.*s\n", Databuf.cchResult, LpchResult);
	    fprintf(FepMsg, "Mode  : %.*s\n", Databuf.cchMode, LpchMode);
	    fprintf(FepMsg, "System: %.*s\n", Databuf.cchSystem, LpchSystem);
	    fprintf(FepMsg, "Buf   : %.*s\n", Databuf.cchBuf, LpchBuf);
	    fprintf(FepMsg, "R:%d M:%d S:%d B:%d C:%d\n",
		    Databuf.cchResult, Databuf.cchMode,
		    Databuf.cchSystem, Databuf.cchBuf, Databuf.cchBufCursor);
	    fflush(FepMsg);
	}
    }
    fflush(FepOut);
}


static void
sendStat(void)
{
    const char	c = FO_Alive;
    write(fileno(FepOut), &c, 1);
    if (DebugFlag) {
	fprintf(FepMsg, "Fep Bridge is Alive\n");
	fflush(FepMsg);
    }
}

static void
sigHandlerFepBrdg(int sig)
{
    terminateFepBrdg();
    exit(0);
}

static int
initFepBrdg(int ac, const char *const av[])
{
    int i;
    const char  *fifoDir = FIFO_DEFAULT_DIR;
    char	*bufFileNameP;
    char        bufPathName[1024];

    DebugFlag = 0;
    for (i = 1; i < ac; i++) {
        if (strcmp("/fifo", av[i]) == 0 || strcmp("/FIFO", av[i]) == 0 
        ||  strcmp("-fifo", av[i]) == 0 || strcmp("-FIFO", av[i]) == 0) {
            i++;
            fifoDir = av[i];
        } else if (strcmp("/debug", av[i]) == 0 || strcmp("/DEBUG", av[i]) == 0
        ||  strcmp("-debug", av[i]) == 0 || strcmp("-DEBUG", av[i]) == 0) {
            DebugFlag = 1;
        } else if (strcmp("/halt", av[i]) == 0 || strcmp("/HALT", av[i]) == 0
        ||  strcmp("-halt", av[i]) == 0 || strcmp("-HALT", av[i]) == 0) {
            HaltFlag = 1;
        } else {
	    usage();
	    return 0;
        }
    }

    i = strlen(fifoDir);
    strcpy(bufPathName, fifoDir);
    if (i > 0
    && (bufPathName[i-1] == ':' || bufPathName[i-1] == '\\')) {
	/* has only drive letter, or ends with '\' */
	bufFileNameP = &bufPathName[i];
    } else {
	bufPathName[i] = '\\';
	bufFileNameP = &bufPathName[i+1];
    }

    printf("%s\n", Version);

    /* Debug Message FIFO open */
    if (DebugFlag) {
        strcpy(bufFileNameP, FIFO_FEP_MSG);
        FepMsg = fopen(bufPathName, "w");
        if (FepMsg == 0) {
            perror(bufPathName);
            return 0;
        }
    } else {
        FepMsg = stdout;        /* stdout */
    }

    /* clear reserved area of Funcbuf, Kkname and Databuf */
    memset(&Funcbuf.Reserved[0], 0, sizeof(Funcbuf.Reserved));
    memset(&Kkbuf.Reserved[0], 0, sizeof(Kkbuf.Reserved));
    memset(&Databuf.Reserved[0], 0, sizeof(Databuf.Reserved));

    if (msknjFind() == 0) {
        fprintf(FepMsg,"MS-KANJI Driver Not Found\n");
	return 0;
    }

    fprintf(FepMsg, "MS-KANJI Driver Found (");
    fprintf(FepMsg, "Name = \"%.8s\" ", Kkbuf.rgchName);
    fprintf(FepMsg, "Version = \"%.4s\" Revision = \"%.4s\")\n",
				Kkbuf.rgchVersion, Kkbuf.rgchRevision);
    fflush(FepMsg);

    /*
	First open FepIn, then FepOut.
	This order of files to open should be the same as on the client
	(Unix) side. If not, the program falls into dead lock !
    */
    strcpy(bufFileNameP, FIFO_FEP_IN);
    FepIn = open(bufPathName, O_RDWR|O_BINARY);
    if (FepIn < 0) {
        perror(bufPathName);
        return 0;
    }
    strcpy(bufFileNameP, FIFO_FEP_OUT);
    FepOut = fopen(bufPathName, "wrb");
    if (FepOut == 0) {
        perror(bufPathName);
        return 0;
    }

    if (msknjOpen() == 0) {
	fprintf(FepMsg, "Cannot open FEP driver.\n");
        return 0;
    }

    /* get information about Result, Mode, System, & Buf */
    CchMaxResult = Databuf.cchResult;
    CchMaxMode = Databuf.cchMode;
    CchMaxSystem = Databuf.cchSystem;
    CchMaxBuf = Databuf.cchBuf;

    if ((LpchResult = (char *)malloc(CchMaxResult)) == 0) {
	fprintf(FepMsg, "Not enough memory.\n");
	return 0;
    }
    Databuf.lpchResult = (char far *)LpchResult;

    if ((LpchMode = (char *)malloc(CchMaxMode)) == 0) {
	fprintf(FepMsg, "Not enough memory.\n");
	return 0;
    }
    Databuf.lpchMode = (char far *)LpchMode;

    if ((LpattrMode = (char *)malloc(CchMaxMode)) == 0) {
	fprintf(FepMsg, "Not enough memory.\n");
	return 0;
    }
    Databuf.lpattrMode = (char far *)LpattrMode;

    if ((LpchSystem = (char *)malloc(CchMaxSystem)) == 0) {
	fprintf(FepMsg, "Not enough memory.\n");
	return 0;
    }
    Databuf.lpchSystem = (char far *)LpchSystem;

    if ((LpattrSystem = (char *)malloc(CchMaxSystem)) == 0) {
	fprintf(FepMsg, "Not enough memory.\n");
	return 0;
    }
    Databuf.lpattrSystem = (char far *)LpattrSystem;

    if ((LpchBuf = (char *)malloc(CchMaxBuf)) == 0) {
	fprintf(FepMsg, "Not enough memory.\n");
	return 0;
    }
    Databuf.lpchBuf = (char far *)LpchBuf;

    if ((LpattrBuf = (char *)malloc(CchMaxBuf)) == 0) {
	fprintf(FepMsg, "Not enough memory.\n");
	return 0;
    }
    Databuf.lpattrBuf = (char far *)LpattrBuf;
    /* Turn Off & On the FEP to reset the FEP */
    msknjSetMode(0);
    msknjSetMode(1);

    /* Store initial Mode & System */
    CchInitMode = Databuf.cchMode;
    if ((LpchInitMode = (char *)malloc(CchInitMode)) == 0) {
	fprintf(FepMsg, "Not enough memory.\n");
	return 0;
    }
    memcpy(LpchInitMode, LpchMode, CchInitMode);
    if ((LpattrInitMode = (char *)malloc(CchInitMode)) == 0) {
	fprintf(FepMsg, "Not enough memory.\n");
	return 0;
    }
    memcpy(LpattrInitMode, LpattrMode, CchInitMode);

    CchInitSystem = Databuf.cchSystem;
    if ((LpchInitSystem = (char *)malloc(CchInitSystem)) == 0) {
	fprintf(FepMsg, "Not enough memory.\n");
	return 0;
    }
    memcpy(LpchInitSystem, LpchSystem, CchInitSystem);
    if ((LpattrInitSystem = (char *)malloc(CchInitSystem)) == 0) {
	fprintf(FepMsg, "Not enough memory.\n");
	return 0;
    }
    memcpy(LpattrInitSystem, LpattrSystem, CchInitSystem);

    return 1;
}

static void
terminateFepBrdg(void)
{
    if (LpchInitMode != 0) {
	free(LpchInitMode);
	/* msknjOpen() have succeeded */
	msknjSetMode(0);
	msknjClose();
    }
    if (LpattrInitMode != 0)	free(LpattrInitMode);
    if (LpchInitSystem != 0)	free(LpchInitSystem);
    if (LpattrInitSystem != 0)	free(LpattrInitSystem);
    if (LpchResult != 0)	free(LpchResult);
    if (LpchMode != 0)		free(LpchMode);
    if (LpattrMode != 0)	free(LpattrMode);
    if (LpchSystem != 0)	free(LpchSystem);
    if (LpattrSystem != 0)	free(LpattrSystem);
    if (LpchBuf != 0)		free(LpchBuf);
    if (LpattrBuf != 0)		free(LpattrBuf);

    LpchResult = LpchMode = LpattrMode = LpchSystem = LpattrSystem
        = LpchBuf = LpattrBuf = LpchInitMode = LpchInitSystem
	= LpattrInitMode = LpattrInitSystem = 0;

    if (FepIn >= 0)  close(FepIn);
    if (FepOut != 0) fclose(FepOut);
    if (FepMsg != 0) fclose(FepMsg);

    fprintf(FepMsg, "FepBrdg terminates.\n");
}

/* inline asm func haltMe() (for LSI C-86) */
static void     _asm_halt(char *asm_statement);
#define haltMe()       _asm_halt("\n\tHLT\n")

static void
haltTheSystem(void)
{
    haltMe();
    /* not reached here */
}

static void
usage(void)
{
    printf(
"%s\n"
"Usage: fepdos [/fifo <fifo_dir>] [/debug] [/halt] [/v]\n"
"Options:\n"
"       /fifo   change the FIFO directory from %s to <fifo_dir>\n"
"       /debug  output debug message to FIFO \"%s\"\n"
"       /halt   halt the system after termination\n"
"       /v      print version and exit the program\n"
    , Version, FIFO_DEFAULT_DIR, FIFO_FEP_MSG);
}

/*
 * Primitive Functions to operate the FEP
 */
static int
msknjFind(void)
{
    union REGS regs;
    struct SREGS sregs;
    int fd, n;

    /*
     * get entry address of KKfunc()
     */
    fd = dopen("MS$KANJI");
    if (fd < 0) {
        fprintf(FepMsg, "MS-KANJI driver not found\n");
        return 0;
    }
    regs.x.ax = RCV_IOC_CHAR;
    regs.x.bx = fd;
    regs.x.cx = 4;
    regs.x.dx = FP_OFF(&KKfunc);
    sregs.ds = FP_SEG(&KKfunc);
    intdosx(&regs, &regs, &sregs);
    n = regs.x.ax;
    dclose(fd);
    if (n != 4) {
        fprintf(FepMsg, "Fail to close the MS-KANJI driver.\n");
        return 0;
    }
    /*
     * check API support level
     */
    Funcbuf.wFunc = KKAsk;
    Funcbuf.wMode = 0;
    Funcbuf.lpKkname = (Kkname far *)&Kkbuf;
    Funcbuf.lpDataparm = 0;
    if(callKKFunc((Funcparm far *)&Funcbuf) < 0){
        fprintf(FepMsg, "Fail to call KKAsk \n");
        return 0;
    }
    if (Kkbuf.wLevel < 1) {
        fprintf(FepMsg, "MS-KANJI driver only supports too low level.\n");
        return 0;
    }
    return 1;
}

static int
msknjOpen(void)
{
    Funcbuf.wFunc = KKOpen;
    Funcbuf.wMode = 0;
    Funcbuf.lpKkname = 0;
    Funcbuf.lpDataparm = (Dataparm far *)&Databuf;
    Databuf.wType = 1;  /* FEP displays nothing */;;
    Databuf.wScan = 0;
    Databuf.wAscii = 0;
    Databuf.wShift = 0;
    Databuf.wExShift = 0;
    if (callKKFunc((Funcparm far *)&Funcbuf) == 0) {
	if (DebugFlag) {
            fprintf(FepMsg, "Size of Result = %d, Mode = %d, ",
                                CchMaxResult, CchMaxMode);
            fprintf(FepMsg, "System = %d, Buf = %d\n",
                                CchMaxSystem,  CchMaxBuf);
            fflush(FepMsg);
        }
    } else {
        fprintf(FepMsg, "MS-KANJI open fail.\n");
        return 0;
    }
    return 1;
}

static int
msknjClose(void)
{
    Funcbuf.wFunc = KKClose;
    Funcbuf.wMode = 0;
    Funcbuf.lpKkname = 0;
    Funcbuf.lpDataparm = 0;
    if( callKKFunc((Funcparm far *)&Funcbuf) == 0 ){
        return 1;
    } else {
        fprintf(FepMsg,"MS-KANJI close fail.\n");
        return 0;
    }
}

static int
msknjInOut(const KeyCode *keyCode)
{
    int ans;
    Funcbuf.wFunc = KKInOut;
    Funcbuf.wMode = 0;
    Funcbuf.lpKkname = 0;
    Funcbuf.lpDataparm = (Dataparm far *)&Databuf;
    Databuf.wType = 0;
    Databuf.wScan = keyCode->scan;
    Databuf.wAscii = keyCode->ascii;
    Databuf.wShift = keyCode->shift;
    Databuf.wExShift = 0;
    if ((ans = callKKFunc((Funcparm far *)&Funcbuf)) < 0) {
        fprintf(FepMsg, "MS-KANJI InOut Fail: ");
        fprintf(FepMsg, "Scan = %02x, Ascii = %02x, ",
                                keyCode->scan, keyCode->ascii);
        fprintf(FepMsg, "Shift = %02x, ExShift = %02x, ",keyCode->shift, 0);
        fflush(FepMsg);
    }
    return ans;
}

static int
msknjGetMode(void)
{
    Funcbuf.wFunc = KKMode;
    Funcbuf.wMode = KM_GET;
    if (callKKFunc((Funcparm far *)&Funcbuf) != 0) {
        fprintf(FepMsg, "MS-KANJI get mode error.\n");
        fflush(FepMsg);
    }
    return (Funcbuf.wMode & KM_OFF)? 0: 1;
}

/*
    This function should call after calling msknjOpen()
*/
static int
msknjSetMode(int mode)
{
    Funcbuf.wFunc = KKMode;
    Funcbuf.wMode = KM_SET;
    Funcbuf.wMode |= (mode == 0)? KM_OFF: KM_ON;
    Funcbuf.wMode |= KM_ECHO;
    Funcbuf.lpKkname = 0;
    Funcbuf.lpDataparm = &Databuf;
    Databuf.wType = 0;
    Databuf.wScan = 0;
    Databuf.wAscii = 0;
    Databuf.wShift = 0;
    Databuf.wExShift = 0;
    if (callKKFunc((Funcparm far *)&Funcbuf) != 0) {
        fprintf(FepMsg, "MS-KANJI get mode error.\n");
        fflush(FepMsg);
    }
    return 0;
}

/* inline asm func sub_sp() (for LSI C-86) */
static void     _asm_sub_sp(char *asm_statement);
#define _q_(s)          # s
#define sub_sp(c)       _asm_sub_sp("\n\tSUB\tSP, " _q_(c) "\n")

static int
callKKFunc(Funcparm far * bufp)
{
    /* 
	We must adjust the stack pointer, because KKfunc() is "Pascal" type.
    */
    (*KKfunc)(bufp);
    sub_sp(4);  	/* Subtract the stack pointer by sizeof(bufp) == 4 */
    /*
	No return statement, because the return value of KKfunc() remains in 
	the AX register, and this will be the return value of this function.
    */
}


/*
        Dos Function Utilities
*/
static int
dopen(const char * dname)
{
    int fd;
    if (_dos_open(dname, O_RDONLY , &fd) != 0) {
        return -1;
    }
    return fd;
}

static void
dclose(int fd)
{
    _dos_close(fd);
}

