/*
  fepbrdg.c : I/F of Fep Bridge 1.34
    Copyright (C) Yutaka Tamiya (tamy@flab.fujitsu.co) 1996/12/15
 */
const char * const Version
    = "Fep Bridge 1.34 by Yutaka Tamiya (tamy@flab.fujitsu.co.jp)";
/*
  FIFO's for communication between DosEmu

  FIFO_DIR	= Location of FIFO's
  FIFO_FEP_IN	= DosEmu's input FIFO
  FIFO_FEP_OUT	= DosEmu's output FIFO
*/
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <netinet/in.h>
#include <netdb.h>
#include <unistd.h>
#include <signal.h>
#include <setjmp.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <assert.h>
#include "fepbrdg.h"

/* Network Service Info */
#ifndef FIFO_DIR
#	define FIFO_DIR	"/var/lib/fepbrdg"
#endif
#define	FIFO_FEP_IN	FIFO_DIR "/" "fepin"
#define	FIFO_FEP_OUT	FIFO_DIR "/" "fepout"
#define	MAXFILE		64
#define DEFAULT_MAX_USER	3
#define	WAKE_UP_INTERVAL_TIME	10L	/* How often to call on DosEmu (sec) */
#ifndef DOSEMU_PATH
#	define DOSEMU_PATH	"/usr/bin/dos"
#endif

/*
  User information
  We distinguish users by their name. Each user has one DosEmu process.
*/
typedef struct {
    char*	name;		/* user name */
    int		nClient;	/* # of client this user has */
    int		dosEmuPid;	/* DosEmu's pid */
    int		fepIn;		/* FIFO's to connect with DosEmu */
    int		fepOut;
} User_t;

/*
  Client information
*/
typedef struct {
    FILE*	fp;		/* client socket */
    User_t*	user;		/* user name */
} Client_t;

/* Global vars. about clients */
static Client_t Client[MAXFILE];		/* array of clients' info */
static volatile int	NClient = 0;		/* # of clients */
static int	MaxClient;			/* limit of # of clients */
static User_t	User[MAXFILE];			/* array of users' info */
static volatile int	NUser = 0;		/* # of users */
static int	MaxUser = DEFAULT_MAX_USER;	/* Max # of users */
static User_t	*WaitUser;			/* user waiting for Dos Emu */

/* Options for DosEmu */
typedef struct {
    const char*	dosEmu;			/* path name of DosEmu */
    const char*	dosEmuConfFile;		/* DosEmu configure file */
    int		debugFlag;		/* flag for debug */
} EmuConf_t;

/* Static functions */
static void  usage(void);
static int   lockPort(const char* hostName);
static int   openServer(int initSock);
static void  mainLoop(int initSock, const EmuConf_t* emuConf);
static int   deleteBadFd(int initSock);
static Client_t*	newClient(int clientSock);
static void  deleteClient(Client_t* client);
static int   serveClient(Client_t* client, const EmuConf_t* emuConf);
static User_t*	userCheck(const char* uname, const EmuConf_t* emuConf);
static int   expandBuf(int len, char** bufP, int* lenBufP);
static void  deleteUser(User_t* user);
static int   startDosEmu(const EmuConf_t* emuConf);
static int   openFepConnection(int* fepIn, int* fepOut);
static int   writeAll(int fd, const void* buf, int count);
static int   fwriteAll(FILE* fp, const void* buf, int count);
static int   readAll(int fd, void* buf, int count);
static int   min(int x, int y);
static int   max(int x, int y);

/* Print error message */
static void  printError(int lineno);
#define	PERROR()	printError(__LINE__)

/*
  main()
*/
int
main(int ac, char* av[])
{
    int		i, rtn = 0;
    const char*	hostName = 0;		/* My host name */
    int		initSock;
    EmuConf_t	emuConf = {DOSEMU_PATH, FIFO_DIR "/dosemu.conf", 0};	

    for (i = 1; i < ac; i++) {
	if (strcmp("-dos", av[i]) == 0) {
	    /* command path should begin with '/' */
	    if (*av[++i] != '/') {
		printf("Error (fepbrdg): Option -dos : "
		       "Path of the dos emulator should begin with /.\n");
		return 0;
	    }
	    emuConf.dosEmu = av[i]; 
	} else if (strcmp("-dosemuconf", av[i]) == 0) {
	    emuConf.dosEmuConfFile = av[++i]; 
	} else if (strcmp("-nodos", av[i]) == 0) {
	    emuConf.dosEmu = 0;
	} else if (strcmp("-user", av[i]) == 0) {
	    MaxUser = max(1, atoi(av[++i])); 
	} else if (strcmp("-hostname", av[i]) == 0) {
	    hostName = av[++i]; 
	} else if (strcmp("-debug", av[i]) == 0) {
	    emuConf.debugFlag = 1;
	} else if (strcmp("-v", av[i]) == 0) {
	    printf("%s\n", Version);
	    return 0;
	} else if (strcmp("-h", av[i]) == 0 || strcmp("-help", av[i]) == 0) {
	    usage();
	    return 0;
	} else if (strcmp("-noauth", av[i]) == 0) {
	    printf("Ignored obsoleted option: -noauth\n");
	} else {
	    usage();
	    return -1;
	}
    }

    /* spawn child process, and the parent process exits */
    if (! emuConf.debugFlag) {
	if (fork() != 0)  _exit(0);
    }

    close(0);	/* not need stdin and stdout */
    close(1);

    /* lock the network port,
	    so that only one "fepbrdg" can work on the same host. */
    if ((initSock = lockPort(hostName)) < 0) { rtn = -1; goto Exit; }
    fprintf(stderr, "\n%s\n", Version);

    /* open server */
    if (openServer(initSock) == 0) { rtn = -1; goto Exit; }

    mainLoop(initSock, &emuConf);

Exit:
    close(initSock);

    return rtn;
}

/*
  Print usage
*/
static void
usage(void)
{
    printf(
"%s\n"
"Usage: fepbrdg [-dos <program>] [-dosemuconf <conf>] [-nodos] \n"
"               [-hostname <name>][-user <num>][-debug] [-v] [-h] [-help]\n"
"Options:\n"
"   -dos         Change DOS Emulator from \"dos\"\n"
"   -dosemuconf  Use DosEmu config file instead of " FIFO_DIR "/dosemu.conf\n"
"   -nodos       Not spawn a new DosEmu process\n"
"   -user        Set maximum # of users, who can use fepbrdg simultaneously.\n"
"                Default is %d\n"
"   -hostname    Manually give this machine name\n"
"   -debug       Usefull for debugging with debuggers\n"
"   -v           Show version\n"
"   -h, -help    Show this message\n"
"\n"	   
"   -noauth      Obsoleted option.\n",
	   Version, DEFAULT_MAX_USER);
}

/*
  Print error message
*/
void
printError(int lineno)
{
    char	buf[256];
    sprintf(buf, "Error (%s: Line %d)", __FILE__, lineno);
    perror(buf);
}

/*
  Lock the Fep Bridge's service port.
*/
static int
lockPort(const char* hostName)
{
    int		sock;		/* init socket */
    const struct servent	*sp;
    int		optbuf = 1;	/* enable socket REUSEADDR */
    struct sockaddr_in	sin;

    if ((sock = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
	PERROR();
	return -1;
    }

    sin.sin_family = AF_INET;
    if (hostName != 0) {
	const struct hostent	*hp;
	if ((hp = gethostbyname(hostName)) == 0) {
	    fprintf(stderr,"Error (fepbrdg: %d): unknown host \"%s\"\n",
			    __LINE__, hostName);
	    goto ErrorExit;
	}
	bcopy(hp->h_addr, &sin.sin_addr, hp->h_length);
    } else {
	sin.sin_addr.s_addr = INADDR_ANY;
    }

    if ((sp = getservbyname(FEP_SERVICE_NAME, "tcp")) != 0) {
	sin.sin_port = sp->s_port;
    } else {
	sin.sin_port = htons((unsigned short)FEP_PORT);
    }

    if (setsockopt(sock, SOL_SOCKET,SO_REUSEADDR,&optbuf,sizeof(optbuf)) < 0) {
	PERROR();
	goto ErrorExit;
    }
    if (bind(sock, (struct sockaddr *)&sin, sizeof(sin)) < 0) {
	PERROR();
	goto ErrorExit;
    }
    return sock;

ErrorExit:
    close(sock);
    return -1;
}

/*
  Open the Fep Bridge's service port.
*/
static int
openServer(int sock)
{
    if (listen(sock, 5) < 0) {
	PERROR();
	return 0;
    }
    return 1;
}

/*
  main loop procedure
      The file descriptor of the larget number should
      be either initSock, fepIn, or fepOut.
*/
static void
childProcessHandler(int sig)
{
    int	status;
    int	pid;
    int	i;
    User_t* user;

    signal(sig, SIG_IGN);
    pid = wait(&status);	/* get the pid of the dead process */
    for (i = 0, user = &User[0]; i < NUser; i++, user++) {
	if (user->dosEmuPid == pid) {
	    Client_t* client;
	    const char* userName = user->name;
	    int	j;
	    /* Prevent deleteClient() do kill(pid) */
	    user->dosEmuPid = -1;	
	    for (j = 0, client = &Client[0]; j < NClient; j++, client++) {
		if (client->user != 0 && client->user->name == userName) {
		    deleteClient(client);
		    j--, client--;
		}
	    }
	    break;
	}
    }
    signal(sig, childProcessHandler);
}

static jmp_buf	EnvOfMainLoop;

/*
  Time out handler in order to jump into the wake-up-routine in mainLoop().
*/
static void
wakeUpHandler(int sig)
{
    signal(sig, wakeUpHandler);
    longjmp(EnvOfMainLoop, 1);
}

/* 
  In order to reduce the network load, we use "buffering" for writing to client
  sockets. We make a write-only stream for each client socket. When we invoke
  fflush() to the client stream, it flushes unwritten data for the client, but
  does not flush unread data from the client. In order to read from the client,
  use the client socket, but not the client stream.
*/
static void
mainLoop(int initSock, const EmuConf_t* emuConf)
{
    int		i;
    User_t*	user;
    Client_t*	client;
    void	(*oldSigChildHandler)(int);
    /* variables for wake up routine */
    const long	wakeUpIntTime
			= (emuConf->debugFlag ? 0L : WAKE_UP_INTERVAL_TIME);
    long	totalElapseTime;		/* total interval time */
    void	(*oldAlarmHandler)(int);	/* old SIGALRM handler */
    static void	wakeUpHandler(int sig);

    /* initialize */
    {
	const int dummyFd = 2;	/* stderr: Any alive fd is OK */
	int maxFd;		/* Max fd ever assigned */
	
	MaxClient = min(MAXFILE, getdtablesize());

	maxFd = initSock;
	for (i = 0, user = &User[0]; i < MaxUser; i++, user++) {
	    user->name = 0;
	    user->nClient = 0;
	    user->dosEmuPid = -1;
	    /* reserve two fd's for each user */
	    user->fepIn  = dup(dummyFd);
	    user->fepOut = dup(dummyFd);
	    maxFd = max(max(maxFd, user->fepOut), user->fepIn);
	}
	MaxClient -= (maxFd + 1);
	if (MaxClient < MaxUser) {
	    /* cannot assign two fifo's and one client for all users */
	    fprintf(stderr,
		"Error (fepbrdg: %d): # of users (%d) is too large.\n",
							__LINE__, MaxUser);
	    return;
	}
    }
    NClient = 0;
    NUser = 0;
    WaitUser = 0;
    oldSigChildHandler = signal(SIGCHLD, childProcessHandler);
    oldAlarmHandler = signal(SIGALRM, wakeUpHandler);
    alarm(wakeUpIntTime);
    totalElapseTime = 0L;

    /* Make sure that main loop to do services for clients
	The file descriptor of the larget number should
	be either initSock, fepIn, or fepOut.
    */
    if (setjmp(EnvOfMainLoop) != 0) {
	/*
	  Wake up routine
	  Wake up the all DosEmu processes sending a meaningless packet.
	  NB: We will arrive here every wakeUpIntTime.
	*/
      WakeUpRoutine:
	for (i = 0, user = &User[0]; i < NUser; i++, user++) {
	    if (user != WaitUser) {
		const char	c = FI_Nop;
		(void) write(user->fepIn, &c, 1);
	    }
	}
	alarm(wakeUpIntTime);	/* alarm restart */
        totalElapseTime = 0L;
    }

    /* main loop */
    while (1) {
	fd_set	readFds;
	long	elapseTime;	/* elapse time after previous call alarm() */

	if (totalElapseTime > wakeUpIntTime) {
	    /* We spent more than wakeUpIntTime without catching SIGALRM's */
	    goto WakeUpRoutine;
	}

	/* Check if any clients want to communicate with me */
	FD_ZERO(&readFds);
	FD_SET(initSock, &readFds);
	for (i = 0, client = &Client[0]; i < NClient; i++, client++) {
	    FD_SET(fileno(client->fp), &readFds);
	}
	/* Also check if any DosEmu processes come ready */
	if (WaitUser != 0) {
	    FD_SET(WaitUser->fepOut, &readFds);
	}
	
	/*
	  Here we've closed stdout and stdin. And we've opened initSock and
	  aissigned two descriptors for each user, where # user >= 1.
	  It's guaranteed that the maximum client's socket < MAXFILE.
	*/
	if (select(MAXFILE, &readFds, NULL, NULL, NULL) < 0) {
	    /* check signal interrupt */
	    switch (errno) {
	    case EINTR: continue;
	    case EBADF:
		if (! deleteBadFd(initSock)) {
		    fprintf(stderr,
			    "Error (fepbrdg: %d) Server socket is down\n",
			    __LINE__);
		    goto Exit;
		}
	    default:
		PERROR();
		continue;
	    }
	}

	/* Check if any DosEmu processes come ready */
        if (WaitUser != 0 && FD_ISSET(WaitUser->fepOut, &readFds)) {
	    char	c;
	    if (readAll(WaitUser->fepOut, &c, 1) == 0 && c == FO_Alive) {
		/* anwer to the waiting client now */
		int	j;
		char	buf[2];
		buf[0] = FO_Ack;
		buf[1] = 0;
		for (j = 0, client = &Client[0]; j < NClient; j++, client++) {
		    if (client->user == WaitUser) {
			fwriteAll(client->fp, buf, 2);
			fflush(client->fp);
		    }
		}
	    } else {
		deleteUser(WaitUser);
	    }
	    WaitUser = 0;	/* clear */
	}

	/* Check request from a new client */
	if (FD_ISSET(initSock, &readFds)) {
	    int	clientSock;		/* a client socket */
	    int	fromlen;
	    struct sockaddr_in	fsin;

	    /* alarm restart */
	    if ((elapseTime = wakeUpIntTime - alarm(wakeUpIntTime)) > 0) {
		totalElapseTime += elapseTime;
	    } else {
		/* the case of less than 1 sec */
		totalElapseTime += 1;
	    }
	    /*
	      I don't disable SIGALRM during the following procedure,
	      because the procedure will end within "wakeUpIntTime"
	    */
	    if ((clientSock
		 = accept(initSock, (struct sockaddr *)&fsin, &fromlen)) < 0) {
		/* fprintf(stderr, "Error (fepbrdg): accept error\n"); */
	    } else {
		newClient(clientSock);
	    }
	}

	/* Check requests from clients */
	for (i = 0, client = &Client[0]; i < NClient; i++, client++) {
	    if (FD_ISSET(fileno(client->fp), &readFds)) {
		/* alarm restart */
		if ((elapseTime = wakeUpIntTime - alarm(wakeUpIntTime)) > 0) {
		    totalElapseTime += elapseTime;
		} else {
		    /* the case of less than 1 sec */
		    totalElapseTime += 1;
		}
	    	/*
	    	  I don't disable SIGALRM during the following procedure,
	    	  because the procedure will end within "wakeUpIntTime"
	    	*/
		switch (serveClient(client, emuConf)) {
		case 1:/* success */
		    break;
		case 0:/* client is dead */
		    deleteClient(client);
		    break;
		case -1:/* Fep Bridge on DosEmu is dead */
		    fprintf(stderr, "Error (fepbrdg: %d): "
				"The DOS emulator seems dead.\n", __LINE__);
		    if (client->user->dosEmuPid >= 0) {
			kill(client->user->dosEmuPid, SIGKILL);
			/* childProcessHandler() will wait() */
		    }
		    break;
		case -2:
		    /* Terminate the program, if we have no users. */
		    if (NUser == 0) {
			goto Exit;
		    } else {
			break;
		    }
		}
	    }
	}
    }
Exit:
    alarm(0);	/* turn off alarm */
    signal(SIGALRM, oldAlarmHandler);
    signal(SIGCHLD, oldSigChildHandler);

    /* close all clients */
    for (i = 0, client = &Client[0]; i < NClient; i++, client++)
	fclose(client->fp);
    for (i = 0, user = &User[0]; i < NUser; i++, user++) {
	int	status;
	free(user->name);
	if (user->dosEmuPid >= 0) {
	    /* send kill code to fepdos.exe */
	    const char	c = FI_Kill;
	    writeAll(user->fepIn, &c, 1);
	    kill(user->dosEmuPid, SIGKILL);	/* force to kill process */
	    wait(&status);
	}
	user->nClient = 0;
	close(user->fepIn);
	close(user->fepOut);
    }
}

/*
  Create a new client. At this point, it has no user information.
*/
static Client_t*
newClient(int clientSock)
{
    Client_t* const	client = &Client[NClient];
    
    assert(clientSock < MAXFILE);
    /* write only stream to prevent fflush() flush input data */
    if ((client->fp = fdopen(clientSock, "w")) == 0 ) {
	fprintf(stderr, "Warning (fepbrdg: %d): max client limit\n", __LINE__);
	close(clientSock);
	return 0;
    }
    if (NClient + 1 == MaxClient) {
	fprintf(stderr, "Warning (fepbrdg: %d): max client limit\n", __LINE__);
	fclose(client->fp);
	return 0;
    }

    client->user = 0;	/* No user information, yet */

    NClient ++;
    return client;
}

/*
  Delete "client". As a result the client's user becomes having no
  clients, also delete the user.
*/ 
static void
deleteClient(Client_t* client)
{
    User_t* const user = client->user;

    /* Delete client from Clinet[] */
    fclose(client->fp);
    NClient -- ;
    *client = Client[NClient];

    if (user == 0) return;

    /* Remove the user, if it has no clients. */
    if (-- user->nClient == 0) {
	/* Remove the user */
	int	status, i;
	    
	free(user->name);
	if (user->dosEmuPid >= 0) {
	    /* disable childProcessHandler */
	    const char	c = FI_Kill;

	    signal(SIGCHLD, SIG_IGN);	

	    /* send kill code to fepdos.exe */
	    writeAll(user->fepIn, &c, 1);
	    kill(user->dosEmuPid, SIGKILL);	/* force to kill process */
	    wait(&status);

	    /* enable childProcessHandler */
	    signal(SIGCHLD, childProcessHandler);
	}

	NUser -- ;
	if (user != &User[NUser]) {
	    /* copy contents of User[NUser] to *user */
	    /* Note that we swap values of fd's, and do not close fd's */
	    int	tmpIn = user->fepIn, tmpOut = user->fepOut;
	    *user = User[NUser];
	    User[NUser].fepIn = tmpIn;
	    User[NUser].fepOut = tmpOut;
	    for (i = 0; i < NClient; i++) {
		if (Client[i].user == &User[NUser]) {
		    Client[i].user = user;
		}
	    }
	}
    }
}

/*
  Find invalide file descriptors among:
  	initSock, client->fp, and user->fepOut,
  then delete them.

  Return 0, if the initSock is dead, otherwise 1.
*/
static int
deleteBadFd(int initSock)
{
    int	i;
    Client_t*	client;
    User_t*	user;
    struct stat	buf;
    
    if (fstat(initSock, &buf) == EBADF) return 0;

    for (i = 0, client = &Client[0]; i < NClient; i++, client++) {
	if (fstat(fileno(client->fp), &buf) == EBADF) {
	    deleteClient(client);
	}
    }
    for (i = 0, user = &User[0]; i < NUser; i++, user++) {
	if (fstat(user->fepOut, &buf) == EBADF) {
	    deleteUser(user);
	}
    }
    return 1;
}

/*
    Service routine for the client
    return  1: Success
    	    0: Client is dead
    	   -1: No responces from fepIn or fepOut (DosEmu program is dead)
    	   -2: catch Kill packet
Note:
    In the case of client is dead, readAll() and writeAll() for the socket
    will fail. In order to read the whole packet comming from FepOut, I have
    to continue calling series of readAll() and writeAll().
*/
static int
serveClient(Client_t* client, const EmuConf_t* emuConf)
{
    char	c;
    FILE *const	clientFp = client->fp;
    const int	clientSock = fileno(client->fp);
    int		fepIn, fepOut;
    int		i, len;
    short	w;

    /* command/resultant buffer for DosEmu */
    static char*	buf = 0;
    static int		lenBuf = 0;

    if (! expandBuf(2 * 5, &buf, &lenBuf)) return -2;

    /* read Packet ID */
    if (readAll(clientSock, &c, 1) < 0) return 0;

    if (client->user != 0) {
	fepIn = client->user->fepIn;
	fepOut = client->user->fepOut;
    }

    switch (c) {
    case FI_Info:
    case FI_Reset:
    case FI_Conv1:
    case FI_ConvN:
	if (client->user == 0) {
	    /* deny access from the client.     */
	    /* The connection should be closed. */
	    return 0;
	}
	if (writeAll(fepIn, &c, 1) < 0) return -1;
	break;
    default:
	break;
    }

    switch (c) {
    case FI_Info:
	    if (readAll(fepOut, &c, 1) < 0) return -1;
	    assert(c == FO_Info);
	    fwriteAll(clientFp, &c, 1);
	    /* ApiLevel, MaxResultStr, MaxModeStr,
					MaxSystemStr, MaxBufStr */
	    if (readAll(fepOut, buf, 2 * 5) < 0) return -1;
	    fwriteAll(clientFp, buf, 2 * 5);

	    /* LenInitMode, LenInitSystem */
	    len = 0;
	    if (readAll(fepOut, &w, 2) < 0) return -1;
	    fwriteAll(clientFp, &w, 2);
	    len += (short)ntohs(w) * 2;
	    if (readAll(fepOut, &w, 2) < 0) return -1;
	    fwriteAll(clientFp, &w, 2);
	    len += (short)ntohs(w) * 2;

	    /* InitModeStr, InitModeAttr, InitSystemStr, InitSystemAttr */
	    if (! expandBuf(len, &buf, &lenBuf)) return -2;
	    if (readAll(fepOut, buf, len) < 0) return -1;
	    if (fwriteAll(clientFp, buf, len) < 0) return 0;
	    fflush(clientFp);
	    break;

    case FI_Auth:
	    /* get user name */
	    if (readAll(clientSock, &w, 2) < 0) return 0;
	    len = (short)ntohs(w);
	    if (! expandBuf(len + 1, &buf, &lenBuf)) return -2;
	    if (readAll(clientSock, buf, len) < 0) return 0;
	    buf[len] = '\0';

	    if (client->user == 0) {
		client->user = userCheck(buf, emuConf);
	    } else {
		/* nop: The client already has sent FI_Auth */
	    }

	    /* New user was denied or DosEmu is ready: Answer to the client */
	    if (client->user == 0 || client->user != WaitUser) {
		buf[0] = FO_Ack;
		buf[1] = (client->user != 0) ? 0 : 1;
		if (fwriteAll(clientFp, buf, 2) < 0) return 0;
		fflush(clientFp);
	    }
	    break;

    case FI_ConvN:
	    /* get N: # of keys will be sent */
	    if (readAll(clientSock, &w, 2) < 0) return 0;
	    if (writeAll(fepIn, &w, 2) < 0) return -1;

	    /* sned N keys to fepIn */
	    for (i = (short)ntohs(w); i > 0; i--) {
		/* scan, ascii, shift */
		if (readAll(clientSock, buf, 3) < 0) {
		    /* send meaningless keycode to fepIn */
		    buf[0] = buf[1] = buf[2] = 0;
		    i = 0;	/* no more iterations */
		}
		if (writeAll(fepIn, buf, 3) < 0) return -1;
	    }
	    goto Reset;
	    /* continue to FI_Reset */

    case FI_Conv1:
	    /* scan, ascii, shift */
	    if (readAll(clientSock, buf, 3) < 0) {
		/* send meaningless keycode to fepIn */
		buf[0] = buf[1] = buf[2] = 0;
	    }
	    if (writeAll(fepIn, buf, 3) < 0) return -1;
	    /* continue to FI_Reset */

    Reset:
    case FI_Reset:
	    if (readAll(fepOut, &c, 1) < 0) return -1;
	    assert(c == FO_Result);
	    fwriteAll(clientFp, &c, 1);

	    /* return value */
	    if (readAll(fepOut, &c, 1) < 0) return -1;
	    fwriteAll(clientFp, &c, 1);
	    if (c == 0) {
		len = 0;
		/* LenResultStr */
		if (readAll(fepOut, &w, 2) < 0) return -1;
		fwriteAll(clientFp, &w, 2);
		len += (short)ntohs(w);

		/* LenModeStr */
		if (readAll(fepOut, &w, 2) < 0) return -1;
		fwriteAll(clientFp, &w, 2);
		len += (short)ntohs(w) * 2;

		/* LenSystemStr */
		if (readAll(fepOut, &w, 2) < 0) return -1;
		fwriteAll(clientFp, &w, 2);
		len += (short)ntohs(w) * 2;

		/* LenBufStr */
		if (readAll(fepOut, &w, 2) < 0) return -1;
		fwriteAll(clientFp, &w, 2);
		len += (short)ntohs(w) * 2;

		/* BufCuror */
		if (readAll(fepOut, &w, 2) < 0) return -1;
		fwriteAll(clientFp, &w, 2);

		/* ResultStr, ModeStr, ModeAttr */
		/* SystemStr, SystemAttr, BufStr, BufAttr */
		if (! expandBuf(len, &buf, &lenBuf)) return -2;
		if (readAll(fepOut, buf, len) < 0) return -1;
		if (fwriteAll(clientFp, buf, len) < 0) return 0;
	    }
	    fflush(clientFp);
	    break;

    case FI_Stat:
	    c = FO_Alive;
	    if (writeAll(clientSock, &c, 1) < 0) return 0;
	    break;

    case FI_UserList:
	    c = FO_UserList;
	    if (writeAll(clientSock, &c, 1) < 0) return 0;
	    w = (short)htons(NUser);
	    if (writeAll(clientSock, &w, 2) < 0) return 0;
	    for (i = 0; i < NUser; i++) {
		len = strlen(User[i].name);
		w = (short)htons(len);
		if (writeAll(clientSock, &w, 2) < 0) return 0;
		if (writeAll(clientSock, User[i].name, len) < 0) return 0;
		w = (short)htons(User[i].nClient);
		if (writeAll(clientSock, &w, 2) < 0) return 0;
	    }
	    break;

    case FI_Kill:
	    /* get user name */
	    if (readAll(clientSock, &w, 2) < 0) return 0;
	    len = (short)ntohs(w);
	    if (! expandBuf(len + 1, &buf, &lenBuf)) return -2;
	    if (readAll(clientSock, buf, len) < 0) return 0;
	    buf[len] = '\0';

	    for (i = NUser - 1; i >= 0; i--) {
		if (strcmp(User[i].name, buf) == 0) break;
	    }
	    if (i >= 0) {
		deleteUser(&User[i]);
	    }
	    return -2;
    }
    return 1;
}

/*
    Expand length of "*bufP" to "len"
*/
static int
expandBuf(int len, char** bufP, int* lenBufP)
{
    if (*lenBufP < len) {
	char*	buf;
	if ((buf = (char *)malloc(len)) == 0) {
	    fprintf(stderr, "Error (fepbrdg: %d): malloc error\n", __LINE__);
	    return 0;
	}
	*lenBufP = len;
	if (*bufP != 0) free(*bufP);
	*bufP = buf;
    }
    return 1;
}

/*
  Check if user "uname" is already entried. If not, create a new user, and
  this user will wait for DosEmu comes.   We set the user to "WaitUser".
  Only one user can wait for DosEmu because more than two DosEMu can share
  FIFO's. 
*/
static User_t*
userCheck(const char* uname, const EmuConf_t* emuConf)
{
    int	i;
    int	dosEmuPid;
    User_t*	user;
    
    for (i = NUser - 1, user = &User[i]; i >= 0; i--, user--) {
	if (strcmp(user->name, uname) == 0) break;
    }
    if (i >= 0) {
	user->nClient ++ ;
	return user;
    } else {
	/* New user */
	char *s;

	if (WaitUser != 0) {
	    /* Another user is waiting for DosEmu */
	    return 0;
	}

	user = &User[NUser];
	if (NUser == MaxUser) return 0;

	if ((s = (char *)malloc(strlen(uname) + 1)) == 0)
	    return 0;
	strcpy(s, uname);

	if (emuConf->dosEmu) {
	    if ((dosEmuPid = startDosEmu(emuConf)) < 0)
		return 0;
	} else {
	    dosEmuPid = -1;
	}

	close(user->fepIn); close(user->fepOut);
	/* open FIFO's : wait until DosEmu process works */
	if (openFepConnection(&user->fepIn, &user->fepOut) == 0) {
	    return 0;
	}
	/*
	fprintf(stderr, "\nfepbrdg: The DOS emulation process starts now.\n");
	*/

	NUser ++;
	user->name = s;
	user->nClient = 1;
	user->dosEmuPid = dosEmuPid;
	WaitUser = user;
	return user;
    }
}

/*
  Delete user and its all clients.
*/
static void
deleteUser(User_t* user)
{
    int j;
    const char* userName = user->name;
    Client_t* client;
    for (j = 0, client = &Client[0]; j < NClient; j++, client++) {
	if (client->user != 0 && client->user->name == userName) {
	    deleteClient(client);
	    j--, client--;
	}
    }
}

/* candidates of the terminal device for DosEmu processes */
static const char *Terminal[] = {
    /* "/dev/tty", */
    "/dev/tty0",
    "/dev/vt00",
    "/dev/systty",
    "/dev/console",
    0
};

#define MaxArgNum	256
#define	MaxCharNum	256

/*
  Start DosEmu as a child process
  return value:
  	pid of child process, if succeeded
	-1, if failed
*/
static int
startDosEmu(const EmuConf_t* emuConf)
{
    int		childPid;
    char*	argv[MaxArgNum];	/* enough number ? */
    char	command[MaxCharNum];	/* enough number ? */
    char	buf1[MaxCharNum];	/* enough number ? */
    enum {DOSEMU, PCEMU, UNKNOWN}	emuType = UNKNOWN;
    int		i;
    int		fd;
    char*	s;
    extern int	OpenPTY(void);

    /* First unlink FIFO's, and make a new FIFO's for a new DosEmu process */
    /* NB: Unlinking FIFO's does not effect on already running DosEmu	   */
    /*     processes, because those FIFO's are still opened by DosEmu	   */
    (void) unlink(FIFO_FEP_IN);
    if (mkfifo(FIFO_FEP_IN, S_IFIFO) != 0 || chmod(FIFO_FEP_IN, 0666) != 0) {
	PERROR();
	return -1;	/* invalid child pid */
    }
    
    (void) unlink(FIFO_FEP_OUT);
    if (mkfifo(FIFO_FEP_OUT, S_IFIFO) != 0 || chmod(FIFO_FEP_OUT, 0666) != 0) {
	PERROR();
	return -1;	/* invalid child pid */
    }

    /* Create a child process, and the parent process returns */
    if ((childPid = vfork()) != 0) {
	if (childPid < 0) {
	    PERROR();
	}
	return childPid;
    }

    /*====================================================*/
    /* The rest of this function is for the child process */
    /*====================================================*/
    /* For pcemu, change the directory to FIFO's dir, which has .pcemurc. */
    chdir(FIFO_DIR);
    
    /* Redirect stdout to /dev/null */
    if (! emuConf->debugFlag) {
	s = "/dev/null";
	if ((fd = open(s, O_WRONLY)) < 0) {
	    fprintf(stderr, "Error (fepbrdg: %d): cannot open \"%s\"\n",
							    __LINE__, s);
	    goto ErrorOfChildProcess;
	}
	close(1); dup(fd);
	close(fd);
    }

    i = 0;
    strcpy(command, emuConf->dosEmu);

    s = strrchr(command, '/');	/* skip directory path */
    argv[i++] = (s != 0) ? (s + 1) : command;
    while ((s = strtok(0, " \t\r\n")) != 0) {
	if (i >= MaxArgNum - 1) goto ErrorOfChildProcess;
	argv[i++] = s;
    }
    if (strcmp(argv[0], "dos") == 0) {
	emuType = DOSEMU;
    } else if (strcmp(argv[0], "pcemu") == 0) {
	emuType = PCEMU;
    }

    switch (emuType) {
    case DOSEMU:
        /* DosEmu 0.60.3 requires that the file descriptor 0 must be:
	    - already open
	    - a terminal device
        */
	close(0);
	if (OpenPTY() < 0) {
	    fprintf(stderr, "Error (fepbrdg: %d): "
	    			"cannot open the terminal.\n", __LINE__);
	    goto ErrorOfChildProcess;
	}
	if (emuConf->dosEmuConfFile != 0) {
	    if (i >= MaxArgNum - 2) goto ErrorOfChildProcess;
	    argv[i++] = "-F";
	    if ((int)strlen(emuConf->dosEmuConfFile) + 1 > MaxCharNum) {
		fprintf(stderr, "Error (fepbrdg: %d): "
			    "too long dos emu config file name.\n", __LINE__);
		goto ErrorOfChildProcess;
	    }
	    argv[i++] = strcpy(buf1, emuConf->dosEmuConfFile);
	}
	break;
    case PCEMU:
	argv[i++] = "-noX";
	break;
    }

    argv[i] = 0;
    if (execv(emuConf->dosEmu, argv) != 0) {
	fprintf(stderr,
		"Error (fepbrdg: %d): cannot execute \"%s\".\n",
					__LINE__, emuConf->dosEmu);
    }

ErrorOfChildProcess:
    kill(0, SIGTERM);	/* kill the parent process */
    exit(1);
}

/*
  Open FIFO's from the parent process.
*/
static int
openFepConnection(int* fepIn, int* fepOut)
{
    int ans = 1;

    /* Open both fifo's in read-write mode, so that *BSD Systems work well. */
    if ((*fepIn = open(FIFO_FEP_IN, O_RDWR)) < 0) {
	fprintf(stderr, "Error (fepbrdg: %d): "
			    "cannot open \"" FIFO_FEP_IN "\"\n", __LINE__);
	ans = 0;
	goto Exit;
    }
    if ((*fepOut = open(FIFO_FEP_OUT, O_RDWR)) < 0) {
	fprintf(stderr, "Error (fepbrdg: %d): "
			    "cannot open \"" FIFO_FEP_OUT "\"\n", __LINE__);
	ans = 0;
	goto Exit;
    }

Exit:
    return ans;
}

static int
min(int x, int y)
{
    return (x < y) ? x : y;
}

static int
max(int x, int y)
{
    return (x > y) ? x : y;
}

/*
  Read data, and never return unless all data of "count" bytes are read.
*/
static int
readAll(int fd, void *buf, int count)
{
    int	n;
    while (count > 0) {
	if ((n = read(fd, buf, count)) <= 0) {
	    if (errno == EINTR) {
		/* Restart, if interrupted by signal */
		continue;
	    } else {
		/* fd is unavailable. The process on the other side is dead.*/
		return -1;
	    }
	}
	buf = (char *)buf + n;
	count -= n;
    }
    return 0;
}

/*
  Write data, and never return unless all data of "count" bytes are written.
*/
static int
writeAll(int fd, const void* buf, int count)
{
    int	n;
    while (count > 0) {
	if ((n = write(fd, buf, count)) < 0) {
	    if (errno == EINTR) {
		/* Restart, if interrupted by signal */
		continue;
	    } else {
		/* fd is unavailable. The process on the other side is dead.*/
		return -1;
	    }
	}
	buf = (const char *)buf + n;
	count -= n;
    }
    return 0;
}

/*
  Write data, and never return unless all data of "count" bytes are written.
*/
static int
fwriteAll(FILE* fp, const void* buf, int count)
{
    int	n;
    while (count > 0) {
	if ((n = fwrite(buf, sizeof(char), count, fp)) < 0) {
	    if (errno == EINTR) {
		/* Restart, if interrupted by signal */
		continue;
	    } else {
		/* fd is unavailable. The process on the other side is dead.*/
		return -1;
	    }
	}
	buf = (const char *)buf + n;
	count -= n;
    }
    return 0;
}
