/*
  feplib.c : I/F of Fep Bridge 1.33
    Copyright (C) Yutaka Tamiya (tamy@flab.fujitsu.co) 1996/11/13
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <netdb.h>
#include <pwd.h>
#include <sys/socket.h>
#include <sys/types.h>

#include "feplib.h"
#include "fepbrdg.h"

extern int SjisToWstr(const char *sjisStr, wchar *result, int flags);
extern int SjisPosToWstrPos(const char *sjisStr, int sjisPos, int flags);
extern int SjisAttrToWstrAttr(const char *sjisStr, char *attr, int flags);

/* local routines */
static int	openConnection(const char *hostName, int noAuth);
static FepOutPacketID	recvPacketID(const FepResult *result);
static int	fepAuthority(int sock);
static void	notConnectedMsg(FepResult *result);
static int	recvInfo(FepResult *result, int convFlags);
static int	recvResult(FepResult *result, int convFlags);
static int	allocBuffer(FepResult *result);
static void	freeBuffer(FepResult *result);

static int	readAll(int fd, void *buf, size_t count);
static int	writeAll(int fd, const void *buf, size_t count);

const char	NotConnectedStr[] = "Not connected to Fep Bridge.";

static char	*RecvBuf = 0;	/* Buffer for FepRecvResult() */
static int	LenRecvBuf = 0;		

#define MAX_LOG_NAME 128

/*
  Export function of feplib
	FepInitialize(), FepTerminate(), FepIsAlive()
	FepInfo(), FepReset(), FepConv(), FepKill()
*/

/*
  Initialize for using Fep Bridge. Open connectin and allocate buffers.
    return value =  1: OK
		 =  0: NG
		 = -1: Not an authorized user
*/
int
FepInitialize(FepResult *result, const char *fepHost, FepConnectMode mode)
{

    const char	*hostName;

    /* First try to get host fepHost, then "$FEPHOST", the local host */
    if ((hostName = fepHost) != 0 || (hostName = getenv("FEPHOST")) != 0) {
	result->HostName = (char *)malloc(strlen(hostName) + 1);
	strcpy(result->HostName, hostName);
    } else {
	char	buf[BUFSIZ];
	if (gethostname(buf, sizeof(buf)) == 0) {
	    result->HostName = (char *)malloc(strlen(buf) + 1);
	    strcpy(result->HostName, buf);
	} else {
	    return 0;
	}
    }

    result->FepSocket = openConnection(result->HostName,
				       mode == FepNoAuth ? 0 : 1);
    switch (result->FepSocket) {
    case -1:
	if (mode != FepWaitForever) {
	    return 0;
	} else {
	    /* wait until Fep Bridge comes */
	    while ((result->FepSocket
		   = openConnection(result->HostName,1)) < 0) {
		sleep(1);	/* wait 1 sec */
	    }
	}
	break;
    case -2:
	return -1;
    }

    /* initialize other fields of result */
    result->ResultStr = result->ModeStr = result->SystemStr = result->BufStr
        = result->InitModeStr = result->InitSystemStr = 0;
    result->ModeAttr = result->SystemAttr = result->BufAttr
        = result->InitModeAttr = result->InitSystemAttr = 0;

    return 1;
}

/*
  Terminate using Fep Bridge. All allocated buffers will be freed.
*/
void
FepTerminate(FepResult *result)
{
    if (result->HostName) free(result->HostName);
    result->HostName = 0;
    freeBuffer(result);
    if (RecvBuf != 0) {
	free(RecvBuf);
	RecvBuf = 0;
	LenRecvBuf = 0;
    }
    close(result->FepSocket);
}

/*
  Ask whether Fep Bridge is alive.
    return value = 1: Alive
		 = 0: Dead
*/
int
FepIsAlive(FepResult *result)
{
    const char	c = FI_Stat;
    FepOutPacketID	id;
    
    writeAll(result->FepSocket, &c, 1);
    id = recvPacketID(result);
    if (id == FO_Alive) {
	return 1;
    } else {
	return 0;
    }
}

/*
  Receive FEP info
    return value = 1: OK
                 = 0: NG
*/
int
FepInfo(FepResult *result, int convFlags)
{
    const char	c = FI_Info;
    int		rtn;

    /* send Request info */
    while (1) {
	if (writeAll(result->FepSocket, &c, 1) >= 0 &&
	(rtn = recvInfo(result, convFlags)) >= 0) {
	    return rtn;
	} else {
	    close(result->FepSocket);
	    while ((result->FepSocket
		   = openConnection(result->HostName, 1)) < 0)
		sleep(1);	/* wait 1sec. */
	}
    }
}

/*
  Reset FEP status
    return value = 1: OK
                 = 0: NG
*/
int
FepReset(FepResult *result, int convFlags)
{
    const char	c = FI_Reset;
    int		i;

    /* send Reset */
    for (i = 0; i < 2; i++) {
	if (i != 0) {
	    close(result->FepSocket);
	    if ((result->FepSocket = openConnection(result->HostName, 1)) < 0) {
		break;
	    }
	}
	if (writeAll(result->FepSocket, &c, 1) >= 0
	&& recvResult(result, convFlags) >= 0) {
	    return 1;
	}
    }
    notConnectedMsg(result);
    return 0;
}

/*
  Kana-Kanji conversion for 1 key.
    return value = 0: OK
                 = 1: FEP did not accept the key.
*/
int
FepConv(FepResult *result, const FepKeyCode *keyCode, int convFlags)
{
    char	buf[4];
    int		rtn;
    int		i;

    buf[0] = FI_Conv;
    buf[1] = keyCode->scan;
    buf[2] = keyCode->ascii;
    buf[3] = keyCode->shift;
    for (i = 0; i < 2; i++) {
	if (i != 0) {
	    close(result->FepSocket);
	    if ((result->FepSocket = openConnection(result->HostName, 1)) < 0) {
		break;
	    }
	}
	if (writeAll(result->FepSocket, buf, 4) >= 0 &&
	(rtn = recvResult(result, convFlags)) >= 0) {
	    return rtn;
	}
    }
    notConnectedMsg(result);
    return 0;
}

static void
notConnectedMsg(FepResult *result)
{
    result->LenModeStr = SjisToWstr(NotConnectedStr, result->ModeStr, 0);
    memset(result->ModeAttr, 0, result->LenModeStr);
    result->LenBufStr = result->LenResultStr = result->LenSystemStr = 0;
}

/*
  Kill Fep Bridge Server
*/
void
FepKill(const FepResult *result)
{
    char 	*buf, tmp[1+2+MAX_LOG_NAME];
    int		len;
    const struct passwd* passwd = getpwuid(getuid());
    
    len = strlen(passwd->pw_name);

    if (len <= MAX_LOG_NAME) {
	buf = tmp;
    } else {
	buf = (char *)malloc(1 + 2 + len);
	if (buf == 0) {
	    return;
	}
    }
    buf[0] = FI_Kill;
    buf[1] = len >> 8;		/* upper byte of len */
    buf[2] = len & 0xff;	/* lower byte of len */
    strcpy(&buf[3], passwd->pw_name);

    writeAll(result->FepSocket, buf, 1 + 2 + len);	/* send Kill */
    if (len > MAX_LOG_NAME) free(buf);
}

void
FepUserList(FepResult *result, FILE* fpOut)
{
    const char	c = FI_UserList;
    FepOutPacketID	id;

    writeAll(result->FepSocket, &c, 1);
    id = recvPacketID(result);
    if (id == FO_UserList) {
	int	nUser, len;
	char	*name = 0;
	int	lenName = 0;
	int	i;
	short	n;

	fprintf(fpOut, "<User>\t\t<Clients>\n");
	/* # User */
	readAll(result->FepSocket, &n, 2);
	nUser = ntohs(n);
	for (i = 0; i < nUser; i++) {
	    /* length of name */
	    readAll(result->FepSocket, &n, 2);
	    len = ntohs(n);
	    if (len > lenName) {
		if (name) free(name);
		if ((name = malloc(len)) == 0) {
		    return;
		}
		lenName = len;
	    }
	    readAll(result->FepSocket, name, len);
	    readAll(result->FepSocket, &n, 2);
	    fprintf(fpOut, "%.*s\t\t%d\n", len, name, ntohs(n));
	}
	if (name) free(name);
    }
}


/*
  return value	=  0: Success
		= -1: Fep Bridge is dead
		= -2: Not an authorized user
*/
static int
openConnection(const char *hostName, int checkAuth)
{
    int		sock;
    const struct hostent	*hp;
    const struct servent	*sp;
    struct sockaddr_in	sin;
    sin.sin_family = AF_INET;

    hp = gethostbyname(hostName);
    if (hp == 0) {
	return -1;
    }
    bcopy(hp->h_addr, &sin.sin_addr, hp->h_length);

    sp = getservbyname(FEP_SERVICE_NAME, "tcp");
    if (sp != 0) {
	sin.sin_port = sp->s_port;
    } else {
	sin.sin_port = htons((unsigned short)FEP_PORT);
    }

    sock = socket(AF_INET, SOCK_STREAM, 0);
    if (sock < 0) {
	return -1;
    }

    if (connect(sock, (struct sockaddr *)&sin, sizeof(sin)) < 0) {
	close(sock);
	return -1;
    }

    if (checkAuth && fepAuthority(sock) == 0) {
	close(sock);
	return -2;
    }

    return sock;
}

/*
  Send client info, and get acknoledge:
    return value = 1: OK
                 = 0: NG
*/
static int
fepAuthority(int sock)
{
    char 	*buf, tmp[1+2+MAX_LOG_NAME];
    int		len;
    int		rtn;
    const struct passwd* passwd = getpwuid(getuid());
    
    len = strlen(passwd->pw_name);

    if (len <= MAX_LOG_NAME) {
	buf = tmp;
    } else {
	buf = (char *)malloc(1 + 2 + len);
	if (buf == 0) {
	    return 0;
	}
    }
    buf[0] = FI_Auth;
    buf[1] = len >> 8;		/* upper byte of len */
    buf[2] = len & 0xff;	/* lower byte of len */
    strcpy(&buf[3], passwd->pw_name);

    if (writeAll(sock, buf, 1 + 2 + len) >= 0) {
	/* wait for FO_Ack packet */
	if (readAll(sock, buf, 2) < 0 || buf[0] != FO_Ack ||  buf[1] != 0) {
	    /* NG to use Fep Bridge */
	    rtn = 0;
	} else {
	    /* OK to use Fep Bridge */
	    rtn = 1;
	}
    } else {
        rtn = 0;
    }
    if (len > MAX_LOG_NAME) free(buf);
    return rtn;
}

static int
recvInfo(FepResult *result, int convFlags)
{
    short	w;
    unsigned short	nw;

    /* wait for FO_Info packet */
    if (recvPacketID(result) != FO_Info) return -1;

    readAll(result->FepSocket, &nw, 2);
    result->ApiLevel = ntohs(nw);

    readAll(result->FepSocket, &nw, 2);
    w = result->MaxResultStr = ntohs(nw);
    if (LenRecvBuf < w) LenRecvBuf = w;

    readAll(result->FepSocket, &nw, 2);
    w = ntohs(nw);
    if (w < sizeof(NotConnectedStr)) w = sizeof(NotConnectedStr);
    result->MaxModeStr = w;
    if (LenRecvBuf < w) LenRecvBuf = w;

    readAll(result->FepSocket, &nw, 2);
    w = result->MaxSystemStr = ntohs(nw);
    if (LenRecvBuf < w) LenRecvBuf = w;

    readAll(result->FepSocket, &nw, 2);
    w = result->MaxBufStr = ntohs(nw);
    if (LenRecvBuf < w) LenRecvBuf = w;

    readAll(result->FepSocket, &nw, 2);
    result->LenInitModeStr = ntohs(nw);

    readAll(result->FepSocket, &nw, 2);
    result->LenInitSystemStr = ntohs(nw);

    /* alloc arrays in FepResult */
    freeBuffer(result);
    if (allocBuffer(result) == 0) {
	return 0;
    }

    /* alloc RecvBuf for FepRecvResult() */
    if (RecvBuf != 0) free(RecvBuf);
    if ((RecvBuf = (char *)malloc(LenRecvBuf)) == 0) {
	return 0;
    }

    /* Store Init Mode String */
    w = result->LenInitModeStr;
    readAll(result->FepSocket, RecvBuf, w);
    RecvBuf[w] = '\0';	/* set terminator */
    result->LenInitModeStr
		    = SjisToWstr(RecvBuf, result->InitModeStr, convFlags);
    readAll(result->FepSocket, result->InitModeAttr, w);
    SjisAttrToWstrAttr(RecvBuf, result->InitModeAttr, convFlags);

    /* Store Init System String */
    w = result->LenInitSystemStr;
    readAll(result->FepSocket, RecvBuf, w);
    RecvBuf[w] = '\0';	/* set terminator */
    result->LenInitSystemStr
		    = SjisToWstr(RecvBuf, result->InitSystemStr, convFlags);
    readAll(result->FepSocket, result->InitSystemAttr, w);
    SjisAttrToWstrAttr(RecvBuf, result->InitSystemAttr, convFlags);

    return 1;
}

static int
recvResult(FepResult *result, int convFlags)
{
    unsigned short	nw;
    short	lenResult, lenMode, lenSystem, lenBuf, bufCursor;
    char	rtn;

    /* wait for FO_Result packet */
    if (recvPacketID(result) != FO_Result) return -1;

    /* Get return value */
    readAll(result->FepSocket, (char *)&rtn, 1);

    if (rtn == 0) {
	/* Lengths of Result, Mode, System, & Buf */
	readAll(result->FepSocket, (char *)&nw, 2);
	lenResult = ntohs(nw);
	readAll(result->FepSocket, (char *)&nw, 2);
	lenMode = ntohs(nw);
	readAll(result->FepSocket, (char *)&nw, 2);
	lenSystem = ntohs(nw);
	readAll(result->FepSocket, (char *)&nw, 2);
	lenBuf = ntohs(nw);

	/* Cursor Buf */
	readAll(result->FepSocket, (char *)&nw, 2);
	bufCursor = ntohs(nw);

	/* Result String */
	readAll(result->FepSocket, RecvBuf, lenResult);
	RecvBuf[lenResult] = '\0';	/* set terminator */
	result->LenResultStr
		= SjisToWstr(RecvBuf, result->ResultStr, convFlags);

	/* Mode String */
	readAll(result->FepSocket, RecvBuf, lenMode);
	RecvBuf[lenMode] = '\0';	/* set terminator */
	result->LenModeStr = SjisToWstr(RecvBuf, result->ModeStr, convFlags);
	readAll(result->FepSocket, result->ModeAttr, lenMode);
	SjisAttrToWstrAttr(RecvBuf, result->ModeAttr, convFlags);

	/* System String */
	readAll(result->FepSocket, RecvBuf, lenSystem);
	RecvBuf[lenSystem] = '\0';	/* set terminator */
	result->LenSystemStr
		= SjisToWstr(RecvBuf, result->SystemStr, convFlags);
	readAll(result->FepSocket, result->SystemAttr, lenSystem);
	SjisAttrToWstrAttr(RecvBuf, result->SystemAttr, convFlags);

	/* Buf String */
	readAll(result->FepSocket, RecvBuf, lenBuf);
	RecvBuf[lenBuf] = '\0';	/* set terminator */
	result->LenBufStr = SjisToWstr(RecvBuf, result->BufStr, convFlags);
	readAll(result->FepSocket, result->BufAttr, lenBuf);
	SjisAttrToWstrAttr(RecvBuf, result->BufAttr, convFlags);

	/* Translate Cursor Buf */
	result->BufCursor = SjisPosToWstrPos(RecvBuf, bufCursor, convFlags);
	    /* Notice: buf contains the SJIS string of BufStr */
    } else {
	result->LenResultStr = 0;
    }

    return rtn;
}

/*
    Sub-routine to recive packet ID
*/
static FepOutPacketID
recvPacketID(const FepResult *result)
{
    char	c;

    /* get the first byte */
    if (readAll(result->FepSocket, &c, 1) < 0) {
	return (FepOutPacketID)(-1);
    }
    return (FepOutPacketID)c;
}

/*
  Sub-routines for FepInitialize() and FepTerminate()
*/
static int
allocBuffer(FepResult *result)
{
    /* In the worst case, the fepsbrdg returns all hanaku-kana strings */
    result->ResultStr
	= (wchar *)malloc((result->MaxResultStr + 1) * sizeof(wchar));
    if (result->ResultStr == 0) {
	return 0;
    }

    result->ModeStr
	= (wchar *)malloc((result->MaxModeStr + 1) * sizeof(wchar));
    if (result->ModeStr == 0) {
	return 0;
    }
    result->ModeAttr
	= (char *)malloc(result->MaxModeStr);
    if (result->ModeAttr == 0) {
	return 0;
    }

    result->SystemStr
	= (wchar *)malloc((result->MaxSystemStr + 1) * sizeof(wchar));
    if (result->SystemStr == 0) {
	return 0;
    }
    result->SystemAttr
	= (char *)malloc(result->MaxSystemStr);
    if (result->SystemAttr == 0) {
	return 0;
    }

    result->BufStr
	= (wchar *)malloc((result->MaxBufStr+1) * sizeof(wchar));
    if (result->BufStr == 0) {
	return 0;
    }
    result->BufAttr
	= (char *)malloc(result->MaxBufStr);
    if (result->BufAttr == 0) {
	return 0;
    }
    result->InitModeStr
	= (wchar *)malloc((result->LenInitModeStr + 1) * sizeof(wchar));
    if (result->InitModeStr == 0) {
	return 0;
    }
    result->InitModeAttr
	= (char *)malloc(result->LenInitModeStr);
    if (result->InitModeAttr == 0) {
	return 0;
    }

    result->InitSystemStr
	= (wchar *)malloc((result->LenInitSystemStr + 1) * sizeof(wchar));
    if (result->InitSystemStr == 0) {
	return 0;
    }
    result->InitSystemAttr
	= (char *)malloc(result->LenInitSystemStr);
    if (result->InitSystemAttr == 0) {
	return 0;
    }
    return 1;
}

static void
freeBuffer(FepResult *result)
{
    if (result->ResultStr != 0)		free(result->ResultStr);
    if (result->ModeStr != 0)		free(result->ModeStr);
    if (result->ModeAttr != 0)		free(result->ModeAttr);
    if (result->SystemStr != 0)		free(result->SystemStr);
    if (result->SystemAttr != 0)	free(result->SystemAttr);
    if (result->BufStr != 0)		free(result->BufStr);
    if (result->BufAttr != 0)		free(result->BufAttr);
    if (result->InitModeStr != 0)	free(result->InitModeStr);
    if (result->InitModeAttr != 0)	free(result->InitModeAttr);
    if (result->InitSystemStr != 0)	free(result->InitSystemStr);
    if (result->InitSystemAttr != 0)	free(result->InitSystemAttr);

    result->ResultStr = result->ModeStr = result->SystemStr = result->BufStr
        = result->InitModeStr = result->InitSystemStr = 0;
    result->ModeAttr = result->SystemAttr = result->BufAttr
        = result->InitModeAttr = result->InitSystemAttr = 0;
}


const int MAX_RETRY = 10000;	/* Retry number to read/write */

static int
readAll(int fd, void *buf, size_t count)
{
    int	n;
    int	retry = MAX_RETRY;
    while (count > 0) {
	n = read(fd, buf, count);
	if (n <= 0 && -- retry == 0) {
	    return -1;
	}
	buf = (char *)buf + n;
	count -= n;
    }
    return 0;
}

static int
writeAll(int fd, const void *buf, size_t count)
{
    int	n;
    int	retry = MAX_RETRY;
    while (count > 0) {
	n = write(fd, buf, count);
	/* if (n < 0) { */
	if (n <= 0 && -- retry == 0) {
	    return -1;
	}
	buf = (const char *)buf + n;
	count -= n;
    }
    return 0;
}
