/*
  sjis2wstr.c : I/F of FEP Bridge 1.33
    Copyright (C) Yutaka Tamiya (tamy@flab.fujitsu.co.jp) 1996/11/13
*/
#include "feplib.h"
#include <stdlib.h>

#define	DEL	0x7f
#define	SP	0x20
#define	SJ0162	0x00e1
#define	SJ6394	0x0161

#define	HankakuKanaDakuten	0xde
#define	HankakuKanaHandakuten	0xdf
#define	wchar_t	wchar

/*
  Export Function
*/
extern int	SjisToWstr(const char* from, wchar *to, int flags);
extern void	SjisAttrToWstrAttr(const char* sjisStr, char *attr, int flags);
extern int	SjisPosToWstrPos(const char* sjisStr, int sjisPos, int flags);

/*
  Static Functions
*/
static void	kanjiSjisToEUC(wchar* buf, int byte1, int byte2);
static void	hankakuKanaToEUC(wchar* buf, int byte, int dakutenFlag);
static wchar	zenkakuAsciiToHankaku(wchar c);
static int	isKanji(const unsigned char* p);
static int	isHankakuKana(const unsigned char* p);

/*
  Convert SJIS string of "from" to Wstr "to"
  Hankaku-Kana characters in "from" are converted to Zenkaku-Kana
  Returns length of "to"
*/
int
SjisToWstr(const char *from, wchar *to, int flags)
{
    int		rtn;
    const unsigned char	*fromP = (const unsigned char *)from;
    wchar	*toP = to;

    for (; *fromP != '\0'; fromP++, toP++) {
	if (isKanji(fromP)) {
	    kanjiSjisToEUC(toP, (int)*fromP, (int)*(fromP+1));
	    if ((flags & FepMaskAsciiZenToHan) != 0) {
		*toP = zenkakuAsciiToHankaku(*toP);
	    }
	    fromP ++;
	} else if ((rtn = isHankakuKana(fromP)) != 0) {
	    /* Hankaku Kana */
	    if ((flags & FepMaskKanaHanToZen) == 0) {
		*toP = (wchar)*fromP;
	    } else {
		hankakuKanaToEUC(toP, (int)*fromP, rtn - 1);
		switch (rtn) {
		case 2:
		case 3: fromP++;
		}
	    }
	} else {
	    /* Ascii */
	    *toP = (wchar)*fromP;
	}
    }
    *toP = (wchar)'\0';
    return toP - to;	/* length of "to" */
}

/*
  Converter: Zenkaku-SJIS code -> EUC
  Returns the next char pointer of the last of "buf".
  I stole this code from:
	    "nkf1.4", written by I. Ichikawa (ichikawa@flab.fujitsu.co.jp)
*/
static void
kanjiSjisToEUC(wchar *buf, int byte1, int byte2)
{
    byte1 = byte1 + byte1 - ((byte1 <= 0x9f) ? SJ0162 : SJ6394);
    if (byte2 < 0x9f)
	byte2 = byte2 - ((byte2 > DEL) ? SP : 0x1f);
    else {
	byte2 = byte2 - 0x7e;
	byte1++;
    }
    *buf = (byte1|0x80) * 256 + (byte2|0x80);
}

/*
  Conversion Table from Hankaku-Kana to EUC Zenkaku-Kana
*/
static const int KanaTable[] = {
    0xa1a3,	/*  */
    0xa1d6,	/*  */
    0xa1d7,	/*  */
    0xa1a2,	/*  */
    0xa1a6,	/*  */
    0xa5f2,	/*  */
    0xa5a1,	/*  */
    0xa5a3,	/*  */
    0xa5a5,	/*  */
    0xa5a7,	/*  */
    0xa5a9,	/*  */
    0xa5e3,	/*  */
    0xa5e5,	/*  */
    0xa5e7,	/*  */
    0xa5c3,	/*  */
    0xa1bc,	/*  */
    0xa5a2,	/*  */
    0xa5a4,	/*  */
    0xa5a6,	/*  */
    0xa5a8,	/*  */
    0xa5aa,	/*  */
    0xa5ab,	/*  */
    0xa5ad,	/*  */
    0xa5af,	/*  */
    0xa5b1,	/*  */
    0xa5b3,	/*  */
    0xa5b5,	/*  */
    0xa5b7,	/*  */
    0xa5b9,	/*  */
    0xa5bb,	/*  */
    0xa5bd,	/*  */
    0xa5bf,	/*  */
    0xa5c1,	/*  */
    0xa5c4,	/*  */
    0xa5c6,	/*  */
    0xa5c8,	/*  */
    0xa5ca,	/*  */
    0xa5cb,	/*  */
    0xa5cc,	/*  */
    0xa5cd,	/*  */
    0xa5ce,	/*  */
    0xa5cf,	/*  */
    0xa5d2,	/*  */
    0xa5d5,	/*  */
    0xa5d8,	/*  */
    0xa5db,	/*  */
    0xa5de,	/*  */
    0xa5df,	/*  */
    0xa5e0,	/*  */
    0xa5e1,	/*  */
    0xa5e2,	/*  */
    0xa5e4,	/*  */
    0xa5e6,	/*  */
    0xa5e8,	/*  */
    0xa5e9,	/*  */
    0xa5ea,	/*  */
    0xa5eb,	/*  */
    0xa5ec,	/*  */
    0xa5ed,	/*  */
    0xa5ef,	/*  */
    0xa5f3,	/*  */
    0xa1ab,	/*  */
    0xa1ac,	/*  */
};

/*
  Convert hankaku-kana of "byte" to EUC zenkaku-kana, and stores in "buf"
  "dakutenFlag" means that "byte" is followed by no dakuten (= 0),
  a Dakuten (= 1), and HanDakuten (= 2).
  Returns the next char pointer of the last of "buf".
*/
static void
hankakuKanaToEUC(wchar *buf, int byte, int dakutenFlag)
{
    *buf = KanaTable[byte - 0xa1] + dakutenFlag; /* 0xa1 == Hankaku " " */
}


/*
  Convert zenkaku-ASCII to hankaku
      This code is contributted by Mr. Sone (ts1@tsn.or.jp)
*/
static struct e2a_struct {
    wchar	euc;
    char	ascii;
} e2a_table[] = {	/* In increasing order of euc value. */
    {0xa1a1, ' '},
    {0xa1a4, ','},
    {0xa1a5, '.'},
    {0xa1a7, ':'},
    {0xa1a8, ';'},
    {0xa1a9, '?'},
    {0xa1aa, '!'},
    {0xa1b0, '^'},
    {0xa1b1, '~'},
    {0xa1b2, '_'},
    {0xa1bf, '/'},
    {0xa1c3, '|'},
    {0xa1c6, '`'},
    {0xa1c7, '\''},
    {0xa1c9, '"'},
    {0xa1ca, '('},
    {0xa1cb, ')'},
    {0xa1ce, '['},
    {0xa1cf, ']'},
    {0xa1d0, '{'},
    {0xa1d1, '}'},
    {0xa1dc, '+'},
    {0xa1dd, '-'},
    {0xa1e1, '='},
    {0xa1e3, '<'},
    {0xa1e4, '>'},
    {0xa1ef, '\\'},
    {0xa1f0, '$'},
    {0xa1f3, '%'},
    {0xa1f4, '#'},
    {0xa1f5, '&'},
    {0xa1f6, '*'},
    {0xa1f7, '@'},
    {0xa3b0, '0'},
    {0xa3b1, '1'},
    {0xa3b2, '2'},
    {0xa3b3, '3'},
    {0xa3b4, '4'},
    {0xa3b5, '5'},
    {0xa3b6, '6'},
    {0xa3b7, '7'},
    {0xa3b8, '8'},
    {0xa3b9, '9'},
    {0xa3c1, 'A'},
    {0xa3c2, 'B'},
    {0xa3c3, 'C'},
    {0xa3c4, 'D'},
    {0xa3c5, 'E'},
    {0xa3c6, 'F'},
    {0xa3c7, 'G'},
    {0xa3c8, 'H'},
    {0xa3c9, 'I'},
    {0xa3ca, 'J'},
    {0xa3cb, 'K'},
    {0xa3cc, 'L'},
    {0xa3cd, 'M'},
    {0xa3ce, 'N'},
    {0xa3cf, 'O'},
    {0xa3d0, 'P'},
    {0xa3d1, 'Q'},
    {0xa3d2, 'R'},
    {0xa3d3, 'S'},
    {0xa3d4, 'T'},
    {0xa3d5, 'U'},
    {0xa3d6, 'V'},
    {0xa3d7, 'W'},
    {0xa3d8, 'X'},
    {0xa3d9, 'Y'},
    {0xa3da, 'Z'},
    {0xa3e1, 'a'},
    {0xa3e2, 'b'},
    {0xa3e3, 'c'},
    {0xa3e4, 'd'},
    {0xa3e5, 'e'},
    {0xa3e6, 'f'},
    {0xa3e7, 'g'},
    {0xa3e8, 'h'},
    {0xa3e9, 'i'},
    {0xa3ea, 'j'},
    {0xa3eb, 'k'},
    {0xa3ec, 'l'},
    {0xa3ed, 'm'},
    {0xa3ee, 'n'},
    {0xa3ef, 'o'},
    {0xa3f0, 'p'},
    {0xa3f1, 'q'},
    {0xa3f2, 'r'},
    {0xa3f3, 's'},
    {0xa3f4, 't'},
    {0xa3f5, 'u'},
    {0xa3f6, 'v'},
    {0xa3f7, 'w'},
    {0xa3f8, 'x'},
    {0xa3f9, 'y'},
    {0xa3fa, 'z'},
};

static int
e2a_comp(const void *a, const void *b)
{
    return (int) ((struct e2a_struct *) a)->euc
	 - (int) ((struct e2a_struct *) b)->euc;
}

static wchar
zenkakuAsciiToHankaku(wchar c)
{
    const struct e2a_struct *match
	= (struct e2a_struct*)bsearch(&c, e2a_table,
			    sizeof(e2a_table) / sizeof(e2a_table[0]),
			    sizeof(e2a_table[0]), e2a_comp);
    if (!match)
	return c;

    return (wchar)match->ascii;
}

/*
  Get a position in a SJIS string to the position in converted EUC string
  Hankaku-Kana characters in "sjisStr" are converted to Zenkaku-Kana
*/
int
SjisPosToWstrPos(const char* sjisStr, int sjisPos, int flags)
{
    int	 i, numHankaku = 0;
    const unsigned char* p = (const unsigned char *)sjisStr;

    for (i = 0; *p != '\0' && i < sjisPos; i ++, p++) {
	if (isKanji(p)
	|| ((flags & FepMaskKanaHanToZen) != 0 && isHankakuKana(p) > 1)) {
	    p++, i++;
	} else {
	    numHankaku ++;
	}
    }
    /*
      In the case of *p == '\0' && i < sjisPos, sjisPos is beyond the
      end of sjisStr. How can I handle this case ?!
    */
    return (sjisPos + numHankaku) / 2;
}

/*
*/
void
SjisAttrToWstrAttr(const char* sjisStr, char* attr, int flags)
{
    const unsigned char* p = (const unsigned char *)sjisStr;
    int	i = 0, j = 0;

    i = 0, j = 0;
    while (*p != '\0') {
	attr[i] = attr[j];
	if (isKanji(p)) {
	    p += 2; i ++; j += 2;
	} else if ((flags & FepMaskKanaHanToZen) != 0 && isHankakuKana(p) > 1) {
	    /* Hankaku Kana with HankakuKanaDakuten or HankakuKanaHandakuten */
	    p += 2, i++, j += 2;
	} else {
	    p += 1, i++, j += 1;
	}
    }
}

/*
  return 1, if *p is the 1st byte of Kanji
         0, otherwise
*/
static int
isKanji(const unsigned char* p)
{
    int	byte1 = (int)*p;
    if ((0x81 <= byte1 && byte1 <= 0x9f)
       || (0xe0 <= byte1 && byte1 <= 0xfc)) {
	/* Kanji */
	return 1;
    }
    return 0;
}

/*
  return 1, if *p is the hankaku-kana, without any (Han)Dakuten.
         2, if *p is the hankaku-kana, and *(p+1) is Dakuten hankaku-kana.
	 3, if *p is the hankaku-kana, and *(p+1) is HanDakuten hankaku-kana.
	 0, otherwise.
*/
static int
isHankakuKana(const unsigned char* p)
{
    int	byte1 = (int)*p;
    if (0xa1 <= byte1 && byte1 <= 0xdf) {
	/* Hankaku Kana */
	int	byte2 = (int)*(p+1);
	switch (byte2) {
	case HankakuKanaDakuten:	return 2;
	case HankakuKanaHandakuten:	return 3;
	default:			return 1;
	}
    }
    return 0;
}
